
from socket import gethostname

from django.conf import settings
from django.utils.translation import ugettext as _

from mezzanine.conf import register_setting


register_setting(
    name="SHOP_CARD_TYPES",
    description="Sequence of available credit card types for payment.",
    editable=False,
    default=("Mastercard", "Visa", "Diners", "Amex"),
)

register_setting(
    name="SHOP_CART_EXPIRY_MINUTES",
    description="Number of minutes of inactivity until carts are abandoned.",
    editable=False,
    default=30,
) 

register_setting(
    name="SHOP_CHECKOUT_ACCOUNT_ENABLED",
    description="If True, users can create a login for the checkout process.",
    editable=False,
    default=True,
)

register_setting(
    name="SHOP_CHECKOUT_ACCOUNT_REQUIRED",
    description="If True, users must create a login for the checkout process.",
    editable=False,
    default=False,
)    

register_setting(
    name="SHOP_CHECKOUT_STEPS_SPLIT",
    description="If True, the checkout process is split into separate "
        "billing/shipping and payment steps.",
    editable=False,
    default=True,
)

register_setting(
    name="SHOP_CHECKOUT_STEPS_CONFIRMATION",
    description="If True, the checkout process has a final confirmation "
        "step before completion.",
    editable=False,
    default=True,
)

register_setting(
    name="SHOP_CURRENCY_LOCALE",
    description="Controls the formatting of monetary values accord to "
        "the locale module in the python standard library.",
    editable=False,
    default="",
)

register_setting(
    name="SHOP_FORCE_HOST",
    description="Host name that the site should always be accessed via that "
        "matches the SSL certificate.",
    editable=True,
    default="",
)

register_setting(
    name="SHOP_FORCE_SSL_VIEWS",
    description="Sequence of view names that will be forced to run over SSL "
        "when SSL_ENABLED is True.",
    editable=False,
    default=("shop_checkout", "shop_complete", "shop_account"),
)

register_setting(
    name="SHOP_OPTION_TYPE_CHOICES",
    description="Sequence of value/name pairs for types of product options, "
        "eg Size, Colour.",
    editable=False,
    default=(
        (1, _("Size")),
        (2, _("Colour")),
    ),
)

register_setting(
    name="SHOP_ORDER_FROM_EMAIL",
    description="Email address that order receipts should be emailed from.",
    editable=True,
    default="do_not_reply@%s" % gethostname(),
)

register_setting(
    name="SHOP_ORDER_STATUS_CHOICES",
    description="Sequence of value/name pairs for order statuses.",
    editable=False,
    default=(
        (1, _("Unprocessed")),
        (2, _("Processed")),
    ),
)

register_setting(
    name="SHOP_PER_PAGE_CATEGORY",
    description="Number of products to display per category page.",
    editable=True,
    default=10,
)

register_setting(
    name="SHOP_PER_PAGE_SEARCH",
    description="Number of products to display per page for search results.",
    editable=True,
    default=10,
)

register_setting(
    name="SHOP_MAX_PAGING_LINKS",
    description="Maximum number of paging links to show.",
    editable=True,
    default=15,
)

register_setting(
    name="SHOP_PRODUCT_SORT_OPTIONS",
    description="Sequence of description/field+direction pairs defining "
        "the options available for sorting a list of products.",
    editable=False,
    default=(
        (_("Relevance"), None),
        (_("Least expensive"), "unit_price"),
        (_("Most expensive"), "-unit_price"),
        (_("Recently added"), "-date_added"),
    ),
)

register_setting(
    name="SHOP_SSL_ENABLED",
    description="If True, users will be automatically redirect to HTTPS "
        "for the checkout process.",
    editable=True,
    default=not settings.DEBUG,
)

# Decorator that wraps the given func in the CallableSetting object that calls 
# the func when it is cast to a string.
callable_setting = lambda func: type("", (), {
    "__str__": lambda self: func(), "__repr__": lambda self: "[dynamic]"})()

@callable_setting
def LOGIN_URL():
    from django.core.urlresolvers import resolve, reverse, Resolver404
    from mezzanine.pages.views import page
    try:
        if resolve(settings.LOGIN_URL)[0] is page:
            raise Resolver404
        return settings.LOGIN_URL
    except Resolver404:
        return reverse("shop_account")

register_setting(
    name="SHOP_LOGIN_URL",
    description="Fall back to shop's login view if the view for LOGIN_URL "
        "hasn't been defined.",
    editable=False,
    default=LOGIN_URL,
)

register_setting(
    name="TEMPLATE_ACCESSIBLE_SETTINGS", 
    description=_("Sequence of setting names available within templates."),
    editable=False,
    default=("SHOP_CHECKOUT_ACCOUNT_ENABLED", "SHOP_CHECKOUT_STEPS_SPLIT", 
        "SHOP_LOGIN_URL", "SHOP_MAX_PAGING_LINKS", 
    ),
    append=True,
)




from copy import copy
from datetime import datetime
from itertools import dropwhile, takewhile
from locale import localeconv
from re import match

from django import forms
from django.forms.models import BaseInlineFormSet, ModelFormMetaclass
from django.contrib.auth import authenticate, login
from django.contrib.auth.models import User
from django.utils.datastructures import SortedDict
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

from mezzanine.conf import settings
from mezzanine.core.templatetags.mezzanine_tags import thumbnail

from cartridge.shop import checkout
from cartridge.shop.models import Product, ProductOption, ProductVariation
from cartridge.shop.models import Cart, Order, DiscountCode
from cartridge.shop.utils import make_choices, set_locale


ADD_PRODUCT_ERRORS = {
    "invalid_options": _("The selected options are currently unavailable."),
    "no_stock": _("The selected options are currently not in stock."),
    "no_stock_quantity": _("The selected quantity is currently unavailable."),
}


def get_add_product_form(product):
    """
    Return the form for adding the given product to the cart or the wishlist. 
    """

    class AddProductForm(forms.Form):

        quantity = forms.IntegerField(min_value=1)
        
        def __init__(self, *args, **kwargs):
            """
            Create each ChoiceField for selecting from the product's variations.
            """
            self._to_cart = kwargs.pop("to_cart", True)
            super(AddProductForm, self).__init__(*args, **kwargs)
            option_names, option_labels = zip(*[(f.name, f.verbose_name) 
                for f in ProductVariation.option_fields()])
            option_values = zip(*product.variations.filter(
                unit_price__isnull=False).values_list(*option_names))
            if option_values:
                for i, name in enumerate(option_names):
                    values = filter(None, set(option_values[i]))
                    if values:
                        field = forms.ChoiceField(label=option_labels[i], 
                                                 choices=make_choices(values))
                        self.fields[name] = field
    
        def clean(self):
            """
            Set the form's selected variation if the selected options and
            quantity are valid.
            """
            options = self.cleaned_data.copy()
            quantity = options.pop("quantity", 0)
            # Ensure the product has a price if adding to cart.
            if self._to_cart:
                options["unit_price__isnull"] = False
            error = None
            try:
                variation = product.variations.get(**options)
            except ProductVariation.DoesNotExist:
                error = "invalid_options"
            else:
                if self._to_cart:
                    if not variation.has_stock():
                        error = "no_stock"
                    elif not variation.has_stock(quantity):
                        error = "no_stock_quantity"
            if error is not None:
                raise forms.ValidationError(ADD_PRODUCT_ERRORS[error])
            self.variation = variation
            return self.cleaned_data

    return AddProductForm


class FormsetForm(object):
    """
    Form mixin that provides template methods for iterating through sets of 
    fields by prefix, single fields and finally remaning fields that haven't 
    been iterated with each fieldset made up from a copy of the original form 
    giving access to as_* methods.
    """

    def _fieldset(self, field_names):
        """
        Return a subset of fields by making a copy of the form containing only 
        the given field names.
        """
        fieldset = copy(self)
        if not hasattr(self, "_fields_done"):
            self._fields_done = []
        fieldset.non_field_errors = lambda *args: None
        field_names = filter(lambda f: f not in self._fields_done, field_names)
        fieldset.fields = SortedDict([(f, self.fields[f]) for f in field_names])
        self._fields_done.extend(field_names)
        return fieldset
        
    def values(self):
        """
        Return pairs of label and value for each field.
        """
        for field in self.fields:
            label = self.fields[field].label
            if label is None:
                label = field[0].upper() + field[1:].replace("_", " ")
            yield (label, self.initial.get(field, self.data.get(field, "")))
        
    def __getattr__(self, name):
        """
        Dynamic fieldset caller - matches requested attribute name against 
        pattern for creating the list of field names to use for the fieldset.
        """
        if name == "errors":
            return None
        filters = (
            ("^other_fields$", lambda: 
                self.fields.keys()),
            ("^(\w*)_fields$", lambda name: 
                [f for f in self.fields.keys() if f.startswith(name)]),
            ("^(\w*)_field$", lambda name: 
                [f for f in self.fields.keys() if f == name]),
            ("^fields_before_(\w*)$", lambda name: 
                takewhile(lambda f: f != name, self.fields.keys())),
            ("^fields_after_(\w*)$", lambda name: 
                list(dropwhile(lambda f: f != name, self.fields.keys()))[1:]),
        )
        for filter_exp, filter_func in filters:
            filter_args = match(filter_exp, name)
            if filter_args is not None:
                return self._fieldset(filter_func(*filter_args.groups()))
        raise AttributeError(name)


class OrderForm(FormsetForm, forms.ModelForm):
    """
    Main Form for the checkout process - ModelForm for Order with extra fields 
    for credit card. Used across each step of the checkout process with fields 
    being hidden across each step where applicable.
    """
    
    step = forms.IntegerField(widget=forms.HiddenInput())
    same_billing_shipping = forms.BooleanField(required=False, initial=True,
        label=_("My delivery details are the same as my billing details"))
    remember = forms.BooleanField(required=False, initial=True,
        label=_("Remember my address for next time"))
    card_name = forms.CharField(label=_("Cardholder name"))
    card_type = forms.ChoiceField(label=_("Type"), 
        choices=make_choices(settings.SHOP_CARD_TYPES))
    card_number = forms.CharField(label=_("Card number"))
    card_expiry_month = forms.ChoiceField(
        choices=make_choices(["%02d" % i for i in range(1, 13)]))
    card_expiry_year = forms.ChoiceField()
    card_ccv = forms.CharField(label="CCV")
    
    class Meta:
        model = Order
        fields = [f.name for f in Order._meta.fields if 
            f.name.startswith("billing_detail") or 
            f.name.startswith("shipping_detail")] + ["additional_instructions",     
            "discount_code"]
            
    def __init__(self, request, step, data=None, initial=None, errors=None):
        """
        Handle setting shipping field values to the same as billing field 
        values in case Javascript is disabled, hiding fields for current step 
        and hiding discount_code field if there are currently no active 
        discount codes.
        """

        # Copy billing fields to shipping fields if "same" checked.
        first = step == checkout.CHECKOUT_STEP_FIRST
        last = step == checkout.CHECKOUT_STEP_LAST
        if (first and data is not None and "same_billing_shipping" in data):
            data = copy(data)
            # Prevent second copy occuring for forcing step below when moving
            # backwards in steps.
            data["step"] = step 
            for field in data:
                billing = field.replace("shipping_detail", "billing_detail")
                if "shipping_detail" in field and billing in data:
                    data[field] = data[billing]

        if initial is not None:
            initial["step"] = step
        # Force the specified step in the posted data - this is required to
        # allow moving backwards in steps. 
        if data is not None and int(data["step"]) != step:
            data = copy(data)
            data["step"] = step
            
        super(OrderForm, self).__init__(data=data, initial=initial)
        self._request = request
        self._checkout_errors = errors

        # Determine which sets of fields to hide for each checkout step.
        hidden = None
        if settings.SHOP_CHECKOUT_STEPS_SPLIT:
            if first:
                # Hide the cc fields for billing/shipping if steps are split.
                hidden = lambda f: f.startswith("card_")
            elif step == checkout.CHECKOUT_STEP_PAYMENT:
                # Hide the non-cc fields for payment if steps are split.
                hidden = lambda f: not f.startswith("card_")
        if settings.SHOP_CHECKOUT_STEPS_CONFIRMATION and last:
            # Hide all fields for the confirmation step.
            hidden = lambda f: True
        if hidden is not None:
            for field in self.fields:
                if hidden(field):
                    self.fields[field].widget = forms.HiddenInput()
                    self.fields[field].required = False
            
        # Hide Discount Code field if no codes are active.
        if DiscountCode.objects.active().count() == 0:
            self.fields["discount_code"].widget = forms.HiddenInput()

        # Set the choices for the cc expiry year relative to the current year.
        year = datetime.now().year
        choices = make_choices(range(year, year + 21))
        self.fields["card_expiry_year"].choices = choices
    
    def clean_discount_code(self):
        """
        validate the discount code if given and update the session with free 
        shipping and/or the discount amount
        """
        code = self.cleaned_data.get("discount_code", "")
        if code:
            cart = Cart.objects.from_request(self._request)
            try:
                discount = DiscountCode.objects.get_valid(code=code, cart=cart)
                self.discount = discount
            except DiscountCode.DoesNotExist:
                error = _("The discount code entered is invalid.")
                raise forms.ValidationError(error)
        return code
        
    def clean(self):
        if self._checkout_errors:
            raise forms.ValidationError(self._checkout_errors)
        return self.cleaned_data


class UserForm(forms.Form):
    """
    Fields for signup & login.
    """
    email = forms.EmailField(label=_("Email Address"))
    password = forms.CharField(label=_("Password"), 
        widget=forms.PasswordInput(render_value=False))

    def authenticate(self):
        """
        Validate email and password as well as setting the user for login.
        """
        self._user = authenticate(username=self.cleaned_data.get("email", ""), 
                               password=self.cleaned_data.get("password", ""))
    
    def login(self, request):
        """
        Log the user in.
        """
        login(request, self._user)


class SignupForm(UserForm):
    
    def clean_email(self):
        """
        Ensure the email address is not already registered.
        """
        email = self.cleaned_data["email"]
        try:
            User.objects.get(username=email)
        except User.DoesNotExist:
            return email
        raise forms.ValidationError(_("This email is already registered"))
        
    def save(self):
        """
        Create the new user using their email address as their username.
        """
        User.objects.create_user(self.cleaned_data["email"], 
            self.cleaned_data["email"], self.cleaned_data["password"])
        self.authenticate()


class LoginForm(UserForm):
    
    def clean(self):
        """
        Authenticate the email/password.
        """
        if "email" in self.cleaned_data and "password" in self.cleaned_data:
            self.authenticate()
            if self._user is None:
                raise forms.ValidationError(_("Invalid email/password"))
            elif not self._user.is_active:
                raise forms.ValidationError(_("Your account is inactive"))
        return self.cleaned_data

    
#######################
#    ADMIN WIDGETS    #    
#######################

class ImageWidget(forms.FileInput):
    """
    Render a visible thumbnail for image fields.
    """
    def render(self, name, value, attrs):
        rendered = super(ImageWidget, self).render(name, value, attrs)
        if value:
            orig_url = "%s%s" % (settings.MEDIA_URL, value)
            thumb_url = "%s%s" % (settings.MEDIA_URL, thumbnail(value, 48, 48))
            rendered = "<a target='_blank' href='%s'><img " \
                "style='margin-right:6px;' src='%s' /></a>%s" % (orig_url, 
                thumb_url, rendered)
        return mark_safe(rendered)


class MoneyWidget(forms.TextInput):
    """
    Render missing decimal places for money fields.
    """
    def render(self, name, value, attrs):
        try:
            value = float(value)
        except (TypeError, ValueError):
            pass
        else:
            set_locale()
            value = ("%%.%sf" % localeconv()["frac_digits"]) % value
            attrs["style"] = "text-align:right;"
        return super(MoneyWidget, self).render(name, value, attrs)


class ProductAdminFormMetaclass(ModelFormMetaclass):
    """
    Metaclass for the Product Admin form that dynamically assigns each of the 
    types of product options as set of checkboxes for selecting which options 
    to use for creating new product variations.
    """
    def __new__(cls, name, bases, attrs):
        for option in settings.SHOP_OPTION_TYPE_CHOICES:
            field = forms.MultipleChoiceField(label=option[1], 
                required=False, widget=forms.CheckboxSelectMultiple)
            attrs["option%s" % option[0]] = field
        args = (cls, name, bases, attrs)
        return super(ProductAdminFormMetaclass, cls).__new__(*args)


class ProductAdminForm(forms.ModelForm):
    """
    Admin form for the Product model.
    """
    __metaclass__ = ProductAdminFormMetaclass
    class Meta:
        model = Product
        
    def __init__(self, *args, **kwargs):
        """
        Set the choices for each of the fields for product options, hiding them 
        if no choices exist.
        """
        super(ProductAdminForm, self).__init__(*args, **kwargs)
        for field, options in ProductOption.objects.as_fields().items():
            self.fields[field].choices = make_choices(options)


class ProductVariationAdminForm(forms.ModelForm):
    """
    Ensure the list of images for the variation are specific to the variation's 
    product.
    """
    def __init__(self, *args, **kwargs):
        super(ProductVariationAdminForm, self).__init__(*args, **kwargs)
        if "instance" in kwargs:
            product = kwargs["instance"].product
            qs = self.fields["image"].queryset.filter(product=product)
            self.fields["image"].queryset = qs


class ProductVariationAdminFormset(BaseInlineFormSet):
    """
    Ensure no more than one variation is checked as default. 
    """
    def clean(self):
        if len([f for f in self.forms if hasattr(f, "cleaned_data") and
            f.cleaned_data["default"]]) > 1:
            error = _("Only one variation can be checked as the default.")
            raise forms.ValidationError(error)


class DiscountAdminForm(forms.ModelForm):
    """
    Ensure only one discount field is given a value and if not, assign the 
    error to the first discount field so that it displays correctly.
    """
    def clean(self):
        fields = [f for f in self.fields if f.startswith("discount_")]
        reductions = filter(None, [self.cleaned_data.get(f) for f in fields])
        if len(reductions) > 1:
            error = _("Please enter a value for only one type of reduction.")
            self._errors[fields[0]] = self.error_class([error])
        return self.cleaned_data




from cartridge.shop.models import Cart


def shop_globals(request):
    """
    Make the cart and wishlist globally available.
    """
    wishlist = request.COOKIES.get("wishlist", "").split(",")
    if not wishlist[0]:
        wishlist = []
    return {"cart": Cart.objects.from_request(request), "wishlist": wishlist}



"""
Various model fields that mostly provide default field sizes to ensure these 
are consistant when used across multiple models.
"""

from locale import localeconv

from django.db.models import CharField, DecimalField
from django.utils.translation import ugettext_lazy as _

from cartridge.shop.utils import set_locale


class OptionField(CharField):
    """
    A field for a selectable option of a product such as colour or size. 
    Ensure Null is True and provide a default field size.
    """
    def __init__(self, *args, **kwargs):
        kwargs["null"] = True
        defaults = {"max_length": 50}
        defaults.update(kwargs)
        super(OptionField, self).__init__(*args, **defaults)

class MoneyField(DecimalField):
    """
    A field for a monetary amount. Provide the default size and precision.
    """
    def __init__(self, *args, **kwargs):
        set_locale()
        defaults = {"null": True, "blank": True, "max_digits": 10, 
            "decimal_places": localeconv()["frac_digits"]}
        defaults.update(kwargs)
        super(MoneyField, self).__init__(*args, **defaults)

class SKUField(CharField):
    """
    A field for a product SKU. Provide the name and default field size.
    """
    def __init__(self, *args, **kwargs):
        if not args:
            args = (_("SKU"),)
        defaults = {"max_length": 20}
        defaults.update(kwargs)
        super(SKUField, self).__init__(*args, **defaults)

class DiscountCodeField(CharField):
    """
    A field for Discount Codes. Provide the default field size.
    """
    def __init__(self, *args, **kwargs):
        defaults = {"max_length": 20}
        defaults.update(kwargs)
        super(DiscountCodeField, self).__init__(*args, **defaults)

# South requires custom fields to be given "rules".
# See http://south.aeracode.org/docs/customfields.html
try:
    from south.modelsinspector import add_introspection_rules
    add_introspection_rules(rules=[((OptionField, MoneyField, SKUField, 
        DiscountCodeField), [], {})], patterns=["cartridge\.shop\.fields\."])
except ImportError:
    pass



#!/usr/bin/env python

"""
Stand-alone data generation routine that uses the ecommerce taxonomy found on 
Google Base to generate a significant amount of category and product data, as 
well as using the Flickr API to retrieve images for the products. The 
multiprocessing module is also used for parallelization.

The Django models and environment used here are specific to the Cartridge 
project but the approach could easily be reused with any ecommerce database.
"""

import sys, os
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), 
    "..", ".."))
os.environ["DJANGO_SETTINGS_MODULE"] = "cartridge.project_template.settings"

from multiprocessing import Process, Queue
from os.path import exists, join
from shutil import move
from sys import exit
from urllib import urlopen, urlretrieve

from django.contrib.webdesign.lorem_ipsum import paragraph
from django.db import connection
from django.db.models import F

from mezzanine.conf import settings
from mezzanine.core.models import CONTENT_STATUS_PUBLISHED

from cartridge.shop.models import Category, Product, ProductOption


try:
    import flickr
except ImportError:
    print "flickr.py must be installed from http://code.google.com/p/flickrpy/"
    exit()

WORKERS = 10
image_dir = join(settings.MEDIA_ROOT, "product")
queue = Queue()
product_options = {"Size": ("Small", "Medium", "Large"), 
    "Colour": ("Red", "Orange", "Yellow", "Green", "Blue", "Indigo", "Violet")}


def create_products(queue):
    """
    Download an image from Flickr for the product on the queue and if 
    successful now or previously, create the applicable product records.
    """

    # Close the connection for this process to avoid the issue discussed here:
    # http://groups.google.com/group/django-users/browse_thread/thread/2c7421cdb9b99e48
    connection.close() 
    product_options = ProductOption.objects.as_fields()
    while True:

        # Get next set of data from queue.
        data = queue.get()
        if data is None:
            break
        main_category, sub_category, product = data[0], data[1], data[-1]

        # Try and download a product image from Flickr.
        image = join(image_dir, "%s.jpg" % product)
        if exists(image):
            message = "Using already downloaded image for %s" % data
        else:
            try:
                images = flickr.photos_search(tags=[product], per_page=1)
                if not images:
                    raise Exception("No images found")
                url = images[0].getURL(size="Large", urlType="source")
                urlretrieve(url, image)
            except Exception, e:
                message = "Error [%s] for %s" % (e, data)
            else:
                message = "Successfully downloaded image for %s" % data
        remaining = "%s remaining" % queue.qsize()
        print remaining.ljust(20, "."), message

        # Create database records for the product.
        if exists(image):
            product = Category.objects.get(parent__title=main_category,
                title=sub_category).products.create(title=product, 
                available=True, status=CONTENT_STATUS_PUBLISHED, 
                content="<p>%s</p>" % paragraph())
            image = "product/%s.jpg" % product.title
            product.images.create(file=image)
            product.variations.create_from_options(product_options)
            product.variations.manage_empty()
            product.variations.update(unit_price=F("id") + "10000")
            product.variations.update(unit_price=F("unit_price") / "1000.0")
            product.copy_default_variation()

if __name__ == "__main__":

    # Load the Google Base data.
    category_url = "http://www.google.com/basepages/producttype/taxonomy.txt"
    try:
        category_data = urlopen(category_url).read()
    except Exception, e:
        print "Failed to load category data: %s" % e
        exit()

    # Clear out the database, moving the product images to a temp location and 
    # restoring them so that they're not deleted.
    print "Resetting product options"
    ProductOption.objects.all().delete()
    for type, name in settings.OPTION_TYPE_CHOICES:
        for name in product_options[unicode(name)]:
            ProductOption.objects.create(type=type, name=name)
        
    Category.objects.all().delete()
    print "Deleting categories"
    Category.objects.all().delete()
    print "Backing up images"
    move(image_dir, "tmp_products")
    print "Deleting products"
    Product.objects.all().delete()
    print "Restoring images"
    move("tmp_products", image_dir)

    # Parse the category data into triples of main category, sub category and 
    # product, create the categories and put the triples onto the queue. The 
    # categories must be created here in a single process due to the non-atomic 
    # nature of Django's Model.objects.get_or_create()
    print "Creating categories"
    for line in category_data.split("\n"):
        parts = line.split(" > ")
        if len(parts) > 2:
            if len(parts) == 3:
                main_category, created = Category.objects.get_or_create(
                    title=parts[0], status=CONTENT_STATUS_PUBLISHED)
                sub_category, created = Category.objects.get_or_create(
                    title=parts[1], status=CONTENT_STATUS_PUBLISHED,
                    parent=main_category)
                queue.put(parts)

    # Create worker processes and run the main function in them.
    workers = []
    for _ in range(WORKERS):
        queue.put(None)
        workers.append(Process(target=create_products, args=(queue,)))
    for worker in workers:
        worker.start()
    for worker in workers:
        worker.join()





from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect, HttpResponsePermanentRedirect

from mezzanine.conf import settings


class SSLRedirect(object):

    def process_request(self, request):
        """
        If SHOP_FORCE_HOST is set and is not the current host, redirect to it
        if SHOP_SSL_ENABLED is True, ensure checkout views are accessed over 
        HTTPS and all other views are accessed over HTTP.
        """
        settings.use_editable()
        force_host = settings.SHOP_FORCE_HOST
        if force_host and request.get_host().split(":")[0] != force_host:
            url = "http://%s%s" % (force_host, request.get_full_path())
            return HttpResponsePermanentRedirect(url)            
        if settings.SHOP_SSL_ENABLED and not settings.DEV_SERVER:
            url = "%s%s" % (request.get_host(), request.get_full_path())
            if request.path in map(reverse, settings.SHOP_FORCE_SSL_VIEWS):
                if not request.is_secure():
                    return HttpResponseRedirect("https://%s" % url)
            elif request.is_secure():
                return HttpResponseRedirect("http://%s" % url)




from django.conf.urls.defaults import *

from mezzanine.conf import settings


urlpatterns = patterns("cartridge.shop.views",
    url("^product/(?P<slug>.*)/$", "product", name="shop_product"),
    url("^search/$", "search", name="shop_search"),
    url("^wishlist/$", "wishlist", name="shop_wishlist"),
    url("^cart/$", "cart", name="shop_cart"),
    url("^checkout/$", "checkout_steps", name="shop_checkout"),
    url("^checkout/complete/$", "complete", name="shop_complete"),
)

if settings.SHOP_CHECKOUT_ACCOUNT_ENABLED:
    urlpatterns += patterns("cartridge.shop.views",
        url("^account/$", "account", name="shop_account"),
        url("^logout/$", "logout", name="shop_logout"),
    )




from django.conf import settings
from django.core.management import call_command
from django.db.models.signals import post_syncdb

from cartridge.shop.models import Product
from cartridge.shop import models as shop_app


def create_initial_product(app, created_models, verbosity, **kwargs):
    if settings.DEBUG and Product in created_models:
        if kwargs.get("interactive"):
            confirm = raw_input("\nWould you like to install an initial "
                                "Category and Product? (yes/no): ")
            while True:
                if confirm == "yes":
                    break
                elif confirm == "no":
                    return
                confirm = raw_input("Please enter either 'yes' or 'no': ")
        print 
        print "Creating initial Category and Product."
        print 
        call_command("loaddata", "cartridge.json")


if "south" not in settings.INSTALLED_APPS:
    post_syncdb.connect(create_initial_product, sender=shop_app)



#!/usr/bin/env python

# When project_template is used as the actual project during Mezzanine
# development, insert the development path into sys.path so that the
# development version of Mezzanine is used rather than the installed version.
import os
import sys
project_path = os.path.dirname(os.path.abspath(__file__))
project_dir = project_path.split(os.sep)[-1]
if project_dir == "project_template":
    dev_path = os.path.abspath(os.path.join(project_path, "..", ".."))
    if dev_path not in sys.path:
        sys.path.insert(0, dev_path)
    import cartridge
    cartridge_path = os.path.dirname(os.path.abspath(cartridge.__file__))
    assert os.path.abspath(os.path.join(cartridge_path, "..")) == dev_path

from django.core.management import execute_manager
try:
    import settings  # Assumed to be in the same directory.
except ImportError:
    import sys
    sys.stderr.write("Error: Can't find the file 'settings.py' in the "
        "directory containing %r. It appears you've customized things.\n"
        "You'll have to run django-admin.py, passing it your settings module.\n"
        "(If the file settings.py does indeed exist, it's causing an "
        "ImportError somehow.)\n" % __file__)
    sys.exit(1)

if __name__ == "__main__":
    execute_manager(settings)




import locale
from urllib import quote

from django import template
from django.template.defaultfilters import slugify

from mezzanine.conf import settings

from cartridge.shop.utils import set_locale


register = template.Library()


@register.filter
def currency(value):
    """
    Format a value as currency according to locale.
    """
    set_locale()
    if len(str(value)) == 0:
        value = 0
    if hasattr(locale, "currency"):
        value = locale.currency(value)
    else:
        # based on locale.currency() in python >= 2.5
        conv = locale.localeconv()
        value = [conv["currency_symbol"], conv["p_sep_by_space"] and " "  or "", 
            (("%%.%sf" % conv["frac_digits"]) % value).replace(".", 
            conv["mon_decimal_point"])]
        if not conv["p_cs_precedes"]:
            value.reverse()
        value = "".join(value)
    return value


def _order_totals(context):
    """
    Add ``item_total``, ``shipping_total``, ``discount_total`` and 
    ``order_total`` to the include context. use the order object for email 
    receipts, or the cart object for checkout.
    """
    if "order" in context:
        context["item_total"] = context["order"].total
        context["shipping_total"] = context["order"].shipping_total
        context["discount_total"] = context["order"].discount_total
    elif "cart" in context:
        context["item_total"] = context["cart"].total_price()
        if context["item_total"] == 0:
            # ignore session if cart has no items as cart may have expired
            # sooner than session
            context["discount_total"] = context["shipping_total"] = 0
        else:
            context["shipping_total"] = context["request"].session.get(
                "shipping_total", None)
            context["discount_total"] = context["request"].session.get(
                "discount_total", None)
    context["order_total"] = context.get("item_total", None)
    if context.get("shipping_total", None) is not None:
        context["order_total"] += context["shipping_total"]
    if context.get("discount_total", None) is not None:
        context["order_total"] -= context["discount_total"]
    return context


@register.inclusion_tag("shop/order_totals.html", takes_context=True)
def order_totals(context):
    """
    HTML version of order_totals.
    """
    return _order_totals(context)

    
@register.inclusion_tag("shop/order_totals.txt", takes_context=True)
def order_totals_text(context):
    """
    Text version of order_totals.
    """
    return _order_totals(context)


@register.inclusion_tag("shop/product_sorting.html", takes_context=True)
def product_sorting(context, products):
    """
    Renders the links for each product sort option.
    """
    sort_options = [(option[0], slugify(option[0])) for option in 
                                        settings.SHOP_PRODUCT_SORT_OPTIONS]
    querystring = context["request"].REQUEST.get("query", "")
    if querystring:
        querystring = "&query=" + quote(querystring)
    else:
        del sort_options[0]
    context.update({"selected_option": getattr(products, "sort"), 
                    "sort_options": sort_options, "querystring": querystring})
    return context


@register.inclusion_tag("shop/product_paging.html", takes_context=True)
def product_paging(context, products):
    """
    Renders the links for each page number in a paginated list of products.
    """
    settings = context["settings"]
    querystring = ""
    for name in ("query", "sort"):
        value = context["request"].REQUEST.get(name)
        if value is not None:
            querystring += "&%s=%s" % (name, quote(value))
    page_range = products.paginator.page_range
    page_links = settings.SHOP_MAX_PAGING_LINKS
    if len(page_range) > page_links:
        start = min(products.paginator.num_pages - page_links, 
            max(0, products.number - (page_links / 2) - 1))
        page_range = page_range[start:start + page_links]
    context.update({"products": products, "querystring": querystring, 
                    "page_range": page_range})
    return context




from django.contrib.auth import logout as auth_logout
from django.core.urlresolvers import reverse
from django.core.paginator import Paginator, InvalidPage, EmptyPage
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.template import RequestContext
from django.template.defaultfilters import slugify
from django.utils import simplejson
from django.utils.translation import ugettext as _

from mezzanine.conf import settings
from mezzanine.utils.views import render_to_response

from cartridge.shop import checkout
from cartridge.shop.forms import OrderForm, LoginForm, SignupForm
from cartridge.shop.forms import get_add_product_form
from cartridge.shop.models import Product, ProductVariation, Cart
from cartridge.shop.utils import set_cookie, sign


# Fall back to authenticated-only messaging if messages app is unavailable.
try:
    from django.contrib.messages import info 
except ImportError:
    def info(request, message, fail_silently=True):
        if request.user.is_authenticated():
            request.user.message_set.create(message=message)
    

def product_list(products, request, per_page):
    """
    Handle pagination and sorting for the given products.
    """
    sort_options = [(slugify(o[0]), o[1]) for o in 
                                        settings.SHOP_PRODUCT_SORT_OPTIONS]
    if "query" not in request.REQUEST:
        del sort_options[0]
    sort_name = request.GET.get("sort", sort_options[0][0])
    sort_value = dict(sort_options).get(sort_name)
    if sort_value is not None:
        products = products.order_by(sort_value)
    paginator = Paginator(products, per_page)
    try:
        page_num = int(request.GET.get("page", 1))
    except ValueError:
        page_num = 1
    try:
        products = paginator.page(page_num)
    except (EmptyPage, InvalidPage):
        products = paginator.page(paginator.num_pages)
    products.sort = sort_name
    return products


def product(request, slug, template="shop/product.html"):
    """
    Display a product - convert the product variations to JSON as well as 
    handling adding the product to either the cart or the wishlist.
    """
    published_products = Product.objects.published(for_user=request.user)
    product = get_object_or_404(published_products, slug=slug)
    AddProductForm = get_add_product_form(product)
    add_product_form = AddProductForm(initial={"quantity": 1})
    if request.method == "POST":
        to_cart = request.POST.get("add_wishlist") is None
        add_product_form = AddProductForm(request.POST, to_cart=to_cart)
        if add_product_form.is_valid():
            if to_cart:
                Cart.objects.from_request(request).add_item(
                    add_product_form.variation, 
                    add_product_form.cleaned_data["quantity"])
                info(request, _("Item added to cart"), fail_silently=True)
                return HttpResponseRedirect(reverse("shop_cart"))
            else:
                skus = request.COOKIES.get("wishlist", "").split(",")
                sku = add_product_form.variation.sku
                if sku not in skus:
                    skus.append(sku)
                info(request, _("Item added to wishlist"), fail_silently=True)
                response = HttpResponseRedirect(reverse("shop_wishlist"))
                set_cookie(response, "wishlist", ",".join(skus))
                return response
    variations = product.variations.all()
    variations_json = simplejson.dumps([dict([(f, getattr(v, f)) for f in 
        ["sku", "image_id"] + [f.name for f in ProductVariation.option_fields()]]) 
        for v in variations])
    return render_to_response(template, {"product": product, "variations_json":
        variations_json, "variations": variations, "images": product.images.all(),
        "add_product_form": add_product_form}, RequestContext(request))


def search(request, template="shop/search_results.html"):
    """
    Display product search results.
    """
    settings.use_editable()
    query = request.REQUEST.get("query", "")
    results = product_list(Product.objects.published_for(user=request.user
                    ).search(query), request, settings.SHOP_PER_PAGE_SEARCH)
    return render_to_response(template, {"query": query, "results": results},
        RequestContext(request))

    
def wishlist(request, template="shop/wishlist.html"):
    """
    Display the wishlist and handle removing items from the wishlist and 
    adding them to the cart.
    """
    skus = request.COOKIES.get("wishlist", "").split(",")
    error = None
    if request.method == "POST":
        sku = request.POST.get("sku")
        to_cart = request.POST.get("add_cart") is not None
        if to_cart:
            quantity = 1
            try:
                variation = ProductVariation.objects.get(sku=sku)
            except ProductVariation:
                error = _("This item is no longer available")
            else:
                if not variation.has_stock(quantity):
                    error = _("This item is currently out of stock")
                else:
                    cart = Cart.objects.from_request(request)
                    cart.add_item(variation, quantity)
        if error is None:
            if sku in skus:
                skus.remove(sku)
            if to_cart:
                message = _("Item add to cart")
                response = HttpResponseRedirect(reverse("shop_cart"))
            else:
                message = _("Item removed from wishlist")
                response = HttpResponseRedirect(reverse("shop_wishlist"))
            info(request, message, fail_silently=True)
            set_cookie(response, "wishlist", ",".join(skus))
            return response
    # Remove skus from the cookie that no longer exist.
    published_products = Product.objects.published(for_user=request.user)
    wishlist = list(ProductVariation.objects.filter(
        product__in=published_products, sku__in=skus).select_related())
    wishlist.sort(key=lambda variation: skus.index(variation.sku))
    response = render_to_response(template, {"wishlist": wishlist, 
        "error": error}, RequestContext(request))
    if len(wishlist) < len(skus):
        skus = [variation.sku for variation in wishlist]
        set_cookie(response, "wishlist", ",".join(skus))
    return response


def cart(request, template="shop/cart.html"):
    """
    Display cart and handle removing items from the cart.
    """
    if request.method == "POST":
        cart = Cart.objects.from_request(request)
        cart.remove_item(request.POST.get("item_id"))
        info(request, _("Item removed from cart"), fail_silently=True)
        return HttpResponseRedirect(reverse("shop_cart"))
    return render_to_response(template, {}, RequestContext(request))


def account(request, template="shop/account.html"):
    """
    Display and handle both the login and signup forms.
    """
    login_form = LoginForm()
    signup_form = SignupForm()
    if request.method == "POST":
        posted_form = None
        message = ""
        if request.POST.get("login") is not None:
            login_form = LoginForm(request.POST)
            if login_form.is_valid():
                posted_form = login_form
                message = _("Successfully logged in")
        else:
            signup_form = SignupForm(request.POST)
            if signup_form.is_valid():
                signup_form.save()
                posted_form = signup_form
                message = _("Successfully signed up")
        if posted_form is not None:
            posted_form.login(request)
            info(request, message, fail_silently=True)
            return HttpResponseRedirect(request.GET.get("next", "/"))
    return render_to_response(template, {"login_form": login_form, 
        "signup_form": signup_form}, RequestContext(request))


def logout(request):
    """
    Log the user out.
    """
    auth_logout(request)
    info(request, _("Successfully logged out"), fail_silently=True)
    return HttpResponseRedirect(request.GET.get("next", "/"))


def checkout_steps(request):
    """
    Display the order form and handle processing of each step.
    """
    
    # Do the authentication check here rather than using standard login_required
    # decorator. This means we can check for a custom LOGIN_URL and fall back
    # to our own login view.
    if settings.SHOP_CHECKOUT_ACCOUNT_REQUIRED and \
        not request.user.is_authenticated():
        return HttpResponseRedirect("%s?next=%s" % (settings.SHOP_LOGIN_URL, 
                                                    reverse("shop_CHECKOUT")))
    
    step = int(request.POST.get("step", checkout.CHECKOUT_STEP_FIRST))
    initial = checkout.initial_order_data(request)
    form = OrderForm(request, step, initial=initial)
    data = request.POST

    if request.POST.get("back") is not None:
        step -= 1
        form = OrderForm(request, step, initial=initial)
    elif request.method == "POST":
        form = OrderForm(request, step, initial=initial, data=data)
        if form.is_valid():
            checkout_errors = []
            request.session["order"] = dict(form.cleaned_data)
            for field in ("card_number", "card_expiry_month", 
                "card_expiry_year", "card_ccv"):
                del request.session["order"][field]

            # Handle shipping and discount code on first step.
            if step == checkout.CHECKOUT_STEP_FIRST:
                try:
                    checkout.billing_shipping(request, form)
                except checkout.CheckoutError, e:
                    checkout_errors.append(e)
                discount = getattr(form, "discount", None)
                if discount is not None:
                    cart = Cart.objects.from_request(request)
                    discount_total = discount.calculate(cart.total_price())
                    request.session["free_shipping"] = discount.free_shipping
                    request.session["discount_total"] = discount_total

            # Process order on final step.
            if step == checkout.CHECKOUT_STEP_LAST and not checkout_errors:
                try:
                    checkout.payment(request, form)
                except checkout.CheckoutError, e:
                    checkout_errors.append(e)
                    if settings.SHOP_CHECKOUT_STEPS_CONFIRMATION:
                        step -= 1
                else:    
                    order = form.save(commit=False)
                    order.process(request)
                    checkout.send_order_email(request, order)
                    response = HttpResponseRedirect(reverse("shop_complete"))
                    if form.cleaned_data.get("remember") is not None:
                        remembered = "%s:%s" % (sign(order.key), order.key)
                        set_cookie(response, "remember", remembered, 
                            secure=request.is_secure())
                    else:
                        response.delete_cookie("remember")
                    return response

            # Assign checkout errors to new form if any and re-run is_valid 
            # if valid set form to next step.
            form = OrderForm(request, step, initial=initial, data=data, 
                             errors=checkout_errors)
            if form.is_valid():
                step += 1
                form = OrderForm(request, step, initial=initial)
            
    template = "shop/%s.html" % checkout.CHECKOUT_TEMPLATES[step - 1]
    return render_to_response(template, {"form": form, "checkout.CHECKOUT_STEP_FIRST": 
        step == checkout.CHECKOUT_STEP_FIRST}, RequestContext(request))


def complete(request, template="shop/complete.html"):
    """
    Redirected to once an order is complete.
    """
    return render_to_response(template, {}, RequestContext(request))




from mezzanine.pages.page_processors import processor_for
from mezzanine.conf import settings

from cartridge.shop.models import Category
from cartridge.shop.views import product_list


@processor_for(Category)
def category_processor(request, page):
    """
    Add paging/sorting to the products for the category.
    """
    settings.use_editable()
    per_page = settings.SHOP_PER_PAGE_CATEGORY
    products = page.category.products.all()
    return {"products": product_list(products, request, per_page)}




from collections import defaultdict
from datetime import datetime, timedelta

from django.db.models import Manager, Q
from django.utils.datastructures import SortedDict

from mezzanine.conf import settings


class CartManager(Manager):

    def from_request(self, request):
        """
        Return a cart by ID stored in the session, creating it if not found
        as well as removing old carts prior to creating a new cart.
        """
        expiry_time = datetime.now() - timedelta(
                                    minutes=settings.SHOP_CART_EXPIRY_MINUTES)
        try:
            cart = self.get(last_updated__gte=expiry_time, 
                id=request.session.get("cart", None))
        except self.model.DoesNotExist:
            self.filter(last_updated__lt=expiry_time).delete()
            cart = self.create()
            request.session["cart"] = cart.id
        else:
            cart.save() # Update timestamp.
        return cart


class ProductOptionManager(Manager):

    def as_fields(self):
        """
        Return a dict of product options as their field names and choices.
        """
        options = defaultdict(list)
        for option in self.all():
            options["option%s" % option.type].append(option.name)
        return options        


class ProductVariationManager(Manager):

    use_for_related_fields = True
    
    def _empty_options_lookup(self, exclude=None):
        """
        Create a lookup dict of field__isnull for options fields.
        """
        if not exclude:
            exclude = {}
        return dict([("%s__isnull" % f.name, True) 
            for f in self.model.option_fields() if f.name not in exclude])

    def create_from_options(self, options):
        """
        Create all unique variations from the selected options.
        """
        if options:
            options = SortedDict(options)
            # Build all combinations of options.
            variations = [[]]
            for values_list in options.values():
                variations = [x + [y] for x in variations for y in values_list]
            for variation in variations:
                # Lookup unspecified options as null to ensure a unique filter.
                variation = dict(zip(options.keys(), variation))
                lookup = dict(variation)
                lookup.update(self._empty_options_lookup(exclude=variation))
                try:
                    self.get(**lookup)
                except self.model.DoesNotExist:
                    self.create(**variation)
                    
    def manage_empty(self):
        """
        Create an empty variation (no options) if none exist, otherwise if 
        multiple variations exist ensure there is no redundant empty variation.
        also ensure there is at least one default variation.
        """
        total_variations = self.count()
        if total_variations == 0:
            self.create()
        elif total_variations > 1:
            self.filter(**self._empty_options_lookup()).delete()
        try:
            self.get(default=True)
        except self.model.DoesNotExist:
            first_variation = self.all()[0]
            first_variation.default = True
            first_variation.save()


class ProductActionManager(Manager):
    
    use_for_related_fields = True

    def _action_for_field(self, field):
        """
        Increases the given field by datetime.today().toordinal() which 
        provides a time scaling value we can order by to determine popularity 
        over time.
        """
        action, created = self.get_or_create(
            timestamp=datetime.today().toordinal())
        setattr(action, field, getattr(action, field) + 1)
        action.save()
    
    def added_to_cart(self):
        """
        Increase total_cart when product is added to cart.
        """
        self._action_for_field("total_cart")

    def purchased(self):
        """
        Increase total_purchased when product is purchased.
        """
        self._action_for_field("total_purchase")


class DiscountCodeManager(Manager):

    def active(self, *args, **kwargs):
        """
        Items flagged as active and in valid date range if date(s) are 
        specified.
        """
        valid_from = Q(valid_from__isnull=True) | Q(valid_from__lte=datetime.now())
        valid_to = Q(valid_to__isnull=True) | Q(valid_to__gte=datetime.now())
        return self.filter(valid_from, valid_to, active=True)
    
    def get_valid(self, code, cart):
        """
        Items flagged as active and within date range as well checking that 
        the given cart contains items that the code is valid for.
        """
        discount = self.active().get(Q(min_purchase__isnull=True) | 
            Q(min_purchase__lte=cart.total_price()), code=code)
        products = discount.products.all()
        if products.count() > 0 and products.filter(variations__sku__in=
            [item.sku for item in cart]).count() == 0:
            raise self.model.DoesNotExist
        return discount




from datetime import datetime
from decimal import Decimal

from django.db import models
from django.db.models.base import ModelBase
from django.utils.translation import ugettext_lazy as _

from mezzanine.conf import settings
from mezzanine.core.models import Displayable, Content
from mezzanine.core.managers import DisplayableManager
from mezzanine.pages.models import Page

from cartridge.shop import fields, managers


class ProductOption(models.Model):
    """
    A selectable option for a product such as size or colour.
    """
    type = models.IntegerField(_("Type"), 
                                    choices=settings.SHOP_OPTION_TYPE_CHOICES)
    name = fields.OptionField(_("Name"))
    
    objects = managers.ProductOptionManager()

    def __unicode__(self):
        return "%s: %s" % (self.get_type_display(), self.name)

    class Meta:
        verbose_name = _("Product option")
        verbose_name_plural = _("Product options")


class Category(Page, Content):
    """
    A category of products on the website.
    """

    class Meta:
        verbose_name = _("Product category")
        verbose_name_plural = _("Product categories")


class Priced(models.Model):
    """
    Abstract model with unit and sale price fields. Inherited by ``Product`` 
    and ``ProductVariation`` models.
    """

    unit_price = fields.MoneyField(_("Unit price"))
    sale_id = models.IntegerField(null=True)
    sale_price = fields.MoneyField(_("Sale price"))
    sale_from = models.DateTimeField(_("Sale start"), blank=True, null=True)
    sale_to = models.DateTimeField(_("Sale end"), blank=True, null=True)

    class Meta:
        abstract = True

    def on_sale(self):
        """
        Returns True if the sale price is applicable.
        """
        valid_from = self.sale_from is None or self.sale_from < datetime.now()
        valid_to = self.sale_to is None or self.sale_to > datetime.now()
        return self.sale_price is not None and valid_from and valid_to

    def has_price(self):
        """
        Returns True if there is a valid price.
        """
        return self.on_sale() or self.unit_price is not None 
    
    def price(self):
        """
        Returns the actual price - sale price if applicable otherwise the unit 
        price.
        """
        if self.on_sale():
            return self.sale_price
        elif self.has_price():
            return self.unit_price
        return Decimal("0")


class Product(Displayable, Priced, Content):
    """
    Container model for a product that stores information common to all of its 
    variations such as the product's title and description.
    """

    available = models.BooleanField(_("Available for purchase"), default=False)
    image = models.CharField(max_length=100, blank=True, null=True)
    categories = models.ManyToManyField("Category", blank=True,
        related_name="products")
    date_added = models.DateTimeField(_("Date added"), auto_now_add=True, 
        null=True)

    objects = DisplayableManager()

    class Meta:
        verbose_name = _("Product")
        verbose_name_plural = _("Products")

    @models.permalink
    def get_absolute_url(self):
        return ("shop_product", (), {"slug": self.slug})

    def copy_default_variation(self):
        """
        Copies the price and image fields from the default variation.
        """
        default = self.variations.get(default=True)
        for field in Priced._meta.fields:
            if not isinstance(field, models.AutoField):
                setattr(self, field.name, getattr(default, field.name))
        if default.image:
            self.image = default.image.file.name
        self.save()

    def admin_thumb(self):
        if self.image is None:
            return ""
        from mezzanine.core.templatetags.mezzanine_tags import thumbnail
        thumb_url = "%s%s" % (settings.MEDIA_URL, thumbnail(self.image, 24, 24))
        return "<img src='%s' />" % thumb_url
    admin_thumb.allow_tags = True
    admin_thumb.short_description = ""


class ProductImage(models.Model):
    """
    An image for a product - a relationship is also defined with the product's 
    variations so that each variation can potentially have it own image, while 
    the relationship between the ``Product`` and ``ProductImage`` models 
    ensures there is a single set of images for the product.
    """
    
    file = models.ImageField(_("Image"), upload_to="product")
    description = models.CharField(_("Description"), blank=True, max_length=100)
    product = models.ForeignKey("Product", related_name="images")
    
    class Meta:
        verbose_name = _("Image")
        verbose_name_plural = _("Images")
    
    def __unicode__(self):
        value = self.description
        if not value:
            value = self.file.name
        if not value:
            value = ""
        return value


class ProductVariationMetaclass(ModelBase):
    """
    Metaclass for the ``ProductVariation`` model that dynamcally assigns an 
    ``fields.OptionField`` for each option in the ``SHOP_PRODUCT_OPTIONS`` 
    setting.
    """
    def __new__(cls, name, bases, attrs):
        for option in settings.SHOP_OPTION_TYPE_CHOICES:
            attrs["option%s" % option[0]] = fields.OptionField(option[1])
        return super(ProductVariationMetaclass, cls).__new__(cls, name, bases, 
            attrs)


class ProductVariation(Priced):
    """
    A combination of selected options from ``SHOP_PRODUCT_OPTIONS`` for a 
    ``Product`` instance.
    """
    
    product = models.ForeignKey("Product", related_name="variations")
    sku = fields.SKUField(unique=True)
    num_in_stock = models.IntegerField(_("Number in stock"), blank=True, 
        null=True) 
    default = models.BooleanField(_("Default"))
    image = models.ForeignKey("ProductImage", null=True, blank=True)

    objects = managers.ProductVariationManager()

    __metaclass__ = ProductVariationMetaclass

    class Meta:
        ordering = ("-default",)
        
    def __unicode__(self):
        """
        Display the option names and values for the variation.
        """
        options = ", ".join(["%s: %s" % (unicode(f.verbose_name), getattr(self, f.name)) 
            for f in self.option_fields() if getattr(self, f.name) is not None])
        return ("%s %s" % (self.product, options)).strip()
    
    def save(self, *args, **kwargs):
        """
        Use the variation's ID as the SKU when the variation is first created 
        and set the variation's image to be the first image of the product if 
        no image is chosen for the variation.
        """
        super(ProductVariation, self).save(*args, **kwargs)
        save = False
        if not self.sku:
            self.sku = self.id
            save = True
        if not self.image:
            image = self.product.images.all()[:1]
            if len(image) == 1:
                self.image = image[0]
                save = True
        if save:
            self.save()

    def get_absolute_url(self):
        return self.product.get_absolute_url()

    @classmethod
    def option_fields(cls):
        return [field for field in cls._meta.fields 
            if isinstance(field, fields.OptionField)]
    
    def options(self):
        return [getattr(self, field.name) for field in self.option_fields()]

    def has_stock(self, quantity=1):
        """
        Check the given quantity is in stock taking into account the number in 
        carts, and cache the number.
        """
        if self.num_in_stock is None or quantity == 0:
            return True
        if not hasattr(self, "_cached_num_in_stock"):
            num_in_stock = self.num_in_stock
            num_in_carts = CartItem.objects.filter(sku=self.sku).aggregate(
                quantity_sum=models.Sum("quantity"))["quantity_sum"]
            if num_in_carts is not None:
                num_in_stock = num_in_stock - num_in_carts
            self._cached_num_in_stock = num_in_stock
        return self._cached_num_in_stock >= quantity


class Order(models.Model):

    billing_detail_first_name = models.CharField(_("First name"), 
        max_length=100)
    billing_detail_last_name = models.CharField(_("Last name"), max_length=100)
    billing_detail_street = models.CharField(_("Street"), max_length=100)
    billing_detail_city = models.CharField(_("City/Suburb"), max_length=100)
    billing_detail_state = models.CharField(_("State/Region"), max_length=100)
    billing_detail_postcode = models.CharField(_("Zip/Postcode"), max_length=10)
    billing_detail_country = models.CharField(_("Country"), max_length=100)
    billing_detail_phone = models.CharField(_("Phone"), max_length=20)
    billing_detail_email = models.EmailField(_("Email"))
    shipping_detail_first_name = models.CharField(_("First name"), 
        max_length=100)
    shipping_detail_last_name = models.CharField(_("Last name"), max_length=100)
    shipping_detail_street = models.CharField(_("Street"), max_length=100)
    shipping_detail_city = models.CharField(_("City/Suburb"), max_length=100)
    shipping_detail_state = models.CharField(_("State/Region"), max_length=100)
    shipping_detail_postcode = models.CharField(_("Zip/Postcode"), 
        max_length=10)
    shipping_detail_country = models.CharField(_("Country"), max_length=100)
    shipping_detail_phone = models.CharField(_("Phone"), max_length=20)
    additional_instructions = models.TextField(_("Additional instructions"), 
        blank=True)
    time = models.DateTimeField(_("Time"), auto_now_add=True, null=True)
    key = models.CharField(max_length=40)
    user_id = models.IntegerField(blank=True, null=True)
    shipping_type = models.CharField(_("Shipping type"), max_length=50, 
        blank=True)
    shipping_total = fields.MoneyField(_("Shipping total"))
    item_total = fields.MoneyField(_("Item total"))
    discount_code = fields.DiscountCodeField(_("Discount code"), blank=True)
    discount_total = fields.MoneyField(_("Discount total"))
    total = fields.MoneyField(_("Order total"))
    status = models.IntegerField(_("Status"), 
        choices=settings.SHOP_ORDER_STATUS_CHOICES, 
        default=settings.SHOP_ORDER_STATUS_CHOICES[0][0])

    class Meta:
        verbose_name = _("Order")
        verbose_name_plural = _("Orders")
        ordering = ("-id",)

    def __unicode__(self):
        return "#%s %s %s" % (self.id, self.billing_name(), self.time)

    def save(self, *args, **kwargs):
        self.total = self.item_total
        if self.shipping_total is not None:
            self.total += self.shipping_total
        if self.discount_total is not None:
            self.total -= self.discount_total
        super(Order, self).save(*args, **kwargs)

    def billing_name(self):
        return "%s %s" % (self.billing_detail_first_name, 
            self.billing_detail_last_name)

    def process(self, request):
        """
        Process a successful order.
        """
        cart = Cart.objects.from_request(request)
        # Get fields fields from session and remove order details from session.
        for field in ("shipping_type", "shipping_total", "discount_total"):
            if field in request.session:
                setattr(self, field, request.session[field])
                del request.session[field]
        del request.session["order"]
        # Set final fields and save.
        self.item_total = cart.total_price()
        self.key = request.session.session_key
        self.user_id = request.user.id    
        self.save()
        # Copy items from cart and delete the cart.
        for item in cart:
            try:
                variation = ProductVariation.objects.get(sku=item.sku)
            except ProductVariation.DoesNotExist:
                pass
            else:
                if variation.num_in_stock is not None:
                    variation.num_in_stock -= item.quantity
                    variation.save()
                variation.product.actions.purchased()
            fields = [f.name for f in SelectedProduct._meta.fields]
            item = dict([(f, getattr(item, f)) for f in fields])
            self.items.create(**item)
        cart.delete()


class Cart(models.Model):

    last_updated = models.DateTimeField(_("Last updated"), auto_now=True, 
        null=True)

    objects = managers.CartManager()

    def __iter__(self):
        """
        Allow the cart to be iterated giving access to the cart's items, 
        ensuring the items are only retrieved once and cached.
        """
        if not hasattr(self, "_cached_items"):
            self._cached_items = self.items.all()
        return iter(self._cached_items)
        
    def add_item(self, variation, quantity):
        """
        Increase quantity of existing item if SKU matches, otherwise create 
        new.
        """
        item, created = self.items.get_or_create(sku=variation.sku, 
            unit_price=variation.price())
        if created:
            item.description = str(variation)
            item.unit_price = variation.price()
            item.url = variation.product.get_absolute_url()
            image = variation.image
            if image is not None:
                item.image = unicode(image.file)
            variation.product.actions.added_to_cart()
        item.quantity += quantity
        item.save()
            
    def remove_item(self, item_id):
        """
        Remove item by SKU.
        """
        try:
            self.items.get(id=item_id).delete()
        except CartItem.DoesNotExist:
            pass
        
    def has_items(self):
        """
        Template helper function - does the cart have items?
        """
        return len(list(self)) > 0 
    
    def total_quantity(self):
        """
        Template helper function - sum of all item quantities?
        """
        return sum([item.quantity for item in self])
        
    def total_price(self):
        """
        Template helper function - sum of all costs of item quantities?
        """
        return sum([item.total_price for item in self])


class SelectedProduct(models.Model):
    """
    Abstract model representing a "selected" product in a cart or order.
    """

    sku = fields.SKUField()
    description = models.CharField(_("Description"), max_length=200)
    quantity = models.IntegerField(_("Quantity"), default=0)
    unit_price = fields.MoneyField(_("Unit price"), default=Decimal("0"))
    total_price = fields.MoneyField(_("Total price"), default=Decimal("0"))
    
    class Meta:
        abstract = True
        
    def __unicode__(self):
        return ""
    
    def save(self, *args, **kwargs):
        self.total_price = self.unit_price * self.quantity
        super(SelectedProduct, self).save(*args, **kwargs)


class CartItem(SelectedProduct):

    cart = models.ForeignKey("Cart", related_name="items")
    url = models.CharField(max_length=200)
    image = models.CharField(max_length=200, null=True)
    
    def get_absolute_url(self):
        return self.url


class OrderItem(SelectedProduct):
    """
    A selected product in a completed order.
    """
    order = models.ForeignKey("Order", related_name="items")


class ProductAction(models.Model):
    """
    Records an incremental value for an action against a product such as adding 
    to cart or purchasing, for sales reporting and calculating popularity. Not 
    yet used but will be used for product popularity and sales reporting.
    """
        
    product = models.ForeignKey("Product", related_name="actions")
    timestamp = models.IntegerField()
    total_cart = models.IntegerField(default=0)
    total_purchase = models.IntegerField(default=0)

    objects = managers.ProductActionManager()

    class Meta:
        unique_together = ("product", "timestamp")


class Discount(models.Model):
    """
    Abstract model representing one of several types of monetary reductions as 
    well as a date range they're applicable for, and the products and products 
    in categories that the reduction is applicable for.
    """

    title = models.CharField(max_length=100)
    active = models.BooleanField(_("Active"))
    products = models.ManyToManyField("Product", blank=True)
    categories = models.ManyToManyField("Category", blank=True)
    discount_deduct = fields.MoneyField(_("Reduce by amount"))
    discount_percent = models.DecimalField(_("Reduce by percent"), max_digits=4, 
        decimal_places=2, blank=True, null=True)
    discount_exact = fields.MoneyField(_("Reduce to amount"))
    valid_from = models.DateTimeField(_("Valid from"), blank=True, null=True)
    valid_to = models.DateTimeField(_("Valid to"), blank=True, null=True)

    class Meta:
        abstract = True

    def __unicode__(self):
        return self.title
        
    def all_products(self):
        """
        Return the selected products as well as the products in the selected 
        categories.
        """
        return Product.objects.filter(models.Q(id__in=self.products.all()) | 
            models.Q(categories__in=self.categories.all()))


class Sale(Discount):
    """
    Stores sales field values for price and date range which when saved are 
    then applied across products and variations according to the selected 
    categories and products for the sale.
    """
    
    class Meta:
        verbose_name = _("Sale")
        verbose_name_plural = _("Sales")
    
    def save(self, *args, **kwargs):
        """
        Apply sales field value to products and variations according to the 
        selected categories and products for the sale.
        """
        super(Sale, self).save(*args, **kwargs)
        self._clear()
        if self.active:
            extra_filter = {}
            if self.discount_deduct is not None:
                # Don't apply to prices that would be negative after deduction.
                extra_filter["unit_price__gt"] = self.discount_deduct
                sale_price = models.F("unit_price") - self.discount_deduct
            elif self.discount_percent is not None:
                sale_price = models.F("unit_price") - (models.F("unit_price") / 
                    "100.0" * self.discount_percent)
            elif self.discount_exact is not None:
                # Don't apply to prices that are cheaper than the sale amount.
                extra_filter["unit_price__gt"] = self.discount_exact
                sale_price = self.discount_exact
            else:
                return
            products = self.all_products()
            variations = ProductVariation.objects.filter(product__in=products)
            for priced_objects in (products, variations):
                # MySQL will raise a 'Data truncated' warning here in some 
                # scenarios, presumably when doing a calculation that exceeds 
                # the precision of the price column. In this case it's safe 
                # to ignore it and the calculation will still be applied.
                try:
                    priced_objects.filter(**extra_filter).update(
                        sale_id=self.id, sale_price=sale_price,
                        sale_to=self.valid_to, sale_from=self.valid_from)
                except Warning:
                    pass
    
    def delete(self, *args, **kwargs):
        self._clear()
        super(Sale, self).delete(*args, **kwargs)

    def _clear(self):
        """
        Clears previously applied sale field values from products prior to 
        updating the sale, when deactivating it or deleting it.
        """
        for priced_model in (Product, ProductVariation):
            priced_model.objects.filter(sale_id=self.id).update(sale_id=None, 
                sale_from=None, sale_to=None, sale_price=None)


class DiscountCode(Discount):
    """
    A code that can be entered at the checkout process to have a discount 
    applied to the total purchase amount.
    """
    
    code = fields.DiscountCodeField(_("Code"), unique=True)
    min_purchase = fields.MoneyField(_("Minimum total purchase"))
    free_shipping = models.BooleanField(_("Free shipping"))

    objects = managers.DiscountCodeManager()
    
    def calculate(self, amount):
        """
        Calculates the discount for the given amount.
        """
        if self.discount_deduct is not None:
            # Don't apply to amounts that would be negative after deduction.
            if self.discount_deduct < amount:
                return self.discount_deduct
        elif self.discount_percent is not None:
            return amount / Decimal("100") * self.discount_percent
        return 0




from django.conf.urls.defaults import *
from django.contrib import admin

from mezzanine.core.views import direct_to_template


admin.autodiscover()

# Add the urlpatterns for any custom Django applications here. 
# You can also change the ``home`` view to add your own functionality to 
# the project's homepage.
urlpatterns = patterns("",
    ("^admin/", include(admin.site.urls)),
    ("^shop/", include("cartridge.shop.urls")),
    url("^$", direct_to_template, {"template": "index.html"}, name="home"),
    ("^", include("mezzanine.urls")),
)



"""

Checkout process utilities. The following are particularly for 
customization and if fail should raise checkout.CheckoutError:

billing_shipping() - Hook for setting shipping via 
``cartridge.shop.utils.set_shipping``.

payment() - Hook for payment gateway integration.

"""

from django.utils.translation import ugettext as _

from mezzanine.conf import settings

from cartridge.shop.models import Order
from cartridge.shop.utils import set_shipping, send_mail_template, sign


class CheckoutError(Exception):
    """
    Raised if an error occurs when processing payment above in 
    shop.checkout.payment() and caught in shop.views.checkout()
    """
    pass


def billing_shipping(request, order_form):
    """
    Implement shipping handling here.
    """
    # Cart is also accessible via shop.Cart.objects.from_request(request)
    set_shipping(request, "Shipping Test", 10)

    
def payment(request, order_form):
    """
    Implement payment gateway integration here.
    """
    # Eg, for declined credit card: raise CheckoutError("Credit card declined")
    pass

    
def initial_order_data(request):
    """
    Return the initial data for the order form - favor request.POST, then 
    session, then last order from either logged in user or from previous order 
    cookie set with "remember my details".
    """
    if request.method == "POST":
        return dict(request.POST.items())
    if "order" in request.session:
        return request.session["order"]
    previous_lookup = {}
    if request.user.is_authenticated():
        previous_lookup["user_id"] = request.user.id
    remembered = request.COOKIES.get("remember", "").split(":")
    if len(remembered) == 2 and remembered[0] == sign(remembered[1]):
        previous_lookup["key"] = remembered[1]
    initial = {}
    if previous_lookup:
        previous_orders = Order.objects.filter(**previous_lookup).values()[:1]
        if len(previous_orders) > 0:
            initial.update(previous_orders[0])
            # Set initial value for "same billing/shipping" based on 
            # whether both sets of address fields are all equal.
            shipping = lambda f: "shipping_%s" % f[len("billing_"):]
            if any([f for f in initial if f.startswith("billing_") and 
                shipping(f) in initial and initial[f] != initial[shipping(f)]]):
                initial["same_billing_shipping"] = False
    return initial

    
def send_order_email(request, order):
    """
    Send order receipt email on successful order.
    """
    settings.use_editable()
    order_context = {"order": order, "request": request, 
        "order_items": order.items.all()}
    for fieldset in ("billing_detail", "shipping_detail"):
        fields = [(f.verbose_name, getattr(order, f.name)) for f in
            order._meta.fields if f.name.startswith(fieldset)]
        order_context["order_%s_fields" % fieldset] = fields
    send_mail_template(_("Order Receipt"), "shop/email/order_receipt", 
        settings.SHOP_ORDER_FROM_EMAIL, order.billing_detail_email, 
        context=order_context)


# Set up some constants for identifying each checkout step.
CHECKOUT_TEMPLATES = ["billing_shipping"]
CHECKOUT_STEP_FIRST = CHECKOUT_STEP_PAYMENT = CHECKOUT_STEP_LAST = 1
if settings.SHOP_CHECKOUT_STEPS_SPLIT:
    CHECKOUT_TEMPLATES.append("payment")
    CHECKOUT_STEP_PAYMENT = CHECKOUT_STEP_LAST = 2
if settings.SHOP_CHECKOUT_STEPS_CONFIRMATION:
    CHECKOUT_TEMPLATES.append("confirmation")
    CHECKOUT_STEP_LAST += 1




import os, sys; sys.path.insert(0, os.path.join("..", "..", "..", "mezzanine"))
from mezzanine.project_template.settings import *

# Cartridge settings.
SHOP_SSL_ENABLED = False

# Mezzanine settings.
from django.utils.translation import ugettext_lazy as _
ADMIN_MENU_ORDER = (
    (_("Content"), ("pages.Page", "blog.BlogPost", "blog.Comment",
        (_("Media Library"), "fb_browse"),)),
    (_("Shop"), ("shop.Product", "shop.ProductOption", "shop.DiscountCode", 
        "shop.Sale", "shop.Order")),
    (_("Site"), ("sites.Site", "redirects.Redirect", "conf.Setting")),
    (_("Users"), ("auth.User", "auth.Group",)),
)

THEME = ""

# Main Django settings.
DEBUG = False
DEV_SERVER = False
MANAGERS = ADMINS = ()
TIME_ZONE = ""
SESSION_EXPIRE_AT_BROWSER_CLOSE = True
LANGUAGE_CODE = "en"
SITE_ID = 1
USE_I18N = False
SECRET_KEY = "%(SECRET_KEY)s"
INTERNAL_IPS = ("127.0.0.1",)

# Databases.
DATABASES = {
    "default": {
        "ENGINE": "",
        "HOST": "",
        "NAME": "",
        "PASSWORD": "",
        "PORT": "",
        "USER": "",
    }
}

# Paths.
import os
_project_path = os.path.dirname(os.path.abspath(__file__))
_project_dir = _project_path.split(os.sep)[-1]
ADMIN_MEDIA_PREFIX = "/media/"
CACHE_MIDDLEWARE_KEY_PREFIX = _project_dir
MEDIA_URL = "/site_media/"
MEDIA_ROOT = os.path.join(_project_path, MEDIA_URL.strip("/"))
ROOT_URLCONF = "%s.urls" % _project_dir
TEMPLATE_DIRS = (os.path.join(_project_path, "templates"),)

# Apps.
INSTALLED_APPS = ("cartridge.shop",) + tuple(INSTALLED_APPS)

TEMPLATE_CONTEXT_PROCESSORS = tuple(TEMPLATE_CONTEXT_PROCESSORS) + (
    "cartridge.shop.context_processors.shop_globals",
)

MIDDLEWARE_CLASSES = tuple(MIDDLEWARE_CLASSES) + (
    "cartridge.shop.middleware.SSLRedirect",
)

# Local settings.
try:
    from local_settings import *
except ImportError:
    pass

# Dynamic settings.
from mezzanine.utils.conf import set_dynamic_settings
set_dynamic_settings(globals())




from __future__ import with_statement
import os

exclude = ["cartridge/project_template/cartridge.db", 
    "cartridge/project_template/local_settings.py"]
exclude = dict([(e, None) for e in exclude])
for e in exclude:
    if e.endswith(".py"):
        try:
            os.remove("%sc" % e)
        except:
            pass
    try:
        with open(e, "r") as f:
            exclude[e] = (f.read(), os.stat(e))
        os.remove(e)
    except:
        pass

from setuptools import setup, find_packages

from cartridge import __version__ as version

try:
    setup(

        name="Cartridge",
        version=version,
        author="Stephen McDonald",
        author_email="stephen.mc@gmail.com",
        description="A Django shopping cart application.",
        long_description=open("README.rst").read(),
        license="BSD",
        url="http://cartridge.jupo.org/",
        zip_safe=False,
        include_package_data=True,
        packages=find_packages(),

        install_requires=[
            "mezzanine >= 0.10",
        ],

        classifiers=[
            "Development Status :: 3 - Alpha",
            "Environment :: Web Environment",
            "Framework :: Django",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: BSD License",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Topic :: Internet :: WWW/HTTP",
            "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
            "Topic :: Internet :: WWW/HTTP :: WSGI",
            "Topic :: Software Development :: Libraries :: "
                                                "Application Frameworks",
            "Topic :: Software Development :: Libraries :: Python Modules",
        ])

finally:
    for e in exclude:
        if exclude[e] is not None:
            data, stat = exclude[e]
            try:
                with open(e, "w") as f:
                    f.write(data)
                os.chown(e, stat.st_uid, stat.st_gid)
                os.chmod(e, stat.st_mode)
            except:
                pass




from copy import deepcopy

from django.contrib import admin
from django.db.models import ImageField
from django.utils.translation import ugettext_lazy as _

from mezzanine.core.admin import DisplayableAdmin, DynamicInlineAdmin
from mezzanine.pages.admin import PageAdmin

from cartridge.shop.fields import MoneyField
from cartridge.shop.forms import ProductAdminForm, ProductVariationAdminForm, \
    ProductVariationAdminFormset, DiscountAdminForm, ImageWidget, MoneyWidget
from cartridge.shop.models import Category, Product, ProductImage, \
    ProductVariation, ProductOption, Order, OrderItem, Sale, DiscountCode


# Lists of field names.
option_fields = [f.name for f in ProductVariation.option_fields()]
billing_fields = [f.name for f in Order._meta.fields 
    if f.name.startswith("billing_detail")]
shipping_fields = [f.name for f in Order._meta.fields 
    if f.name.startswith("shipping_detail")]

category_fieldsets = deepcopy(PageAdmin.fieldsets)
category_fieldsets[0][1]["fields"].insert(3, "content")


class CategoryAdmin(PageAdmin):
    fieldsets = category_fieldsets
    formfield_overrides = {ImageField: {"widget": ImageWidget}}


class ProductVariationAdmin(admin.TabularInline):
    verbose_name_plural = _("Current variations")
    model = ProductVariation
    fields = ("sku", "default", "num_in_stock", "unit_price", "sale_price", 
        "sale_from", "sale_to", "image")
    extra = 0
    formfield_overrides = {MoneyField: {"widget": MoneyWidget}}
    form = ProductVariationAdminForm
    formset = ProductVariationAdminFormset


class ProductImageAdmin(DynamicInlineAdmin):
    model = ProductImage
    formfield_overrides = {ImageField: {"widget": ImageWidget}}
    

product_fieldsets = deepcopy(DisplayableAdmin.fieldsets)
product_fieldsets[0][1]["fields"].extend(["available", "categories", 
    "content"])
product_fieldsets = list(product_fieldsets)
product_fieldsets.insert(1, (_("Create new variations"), 
    {"classes": ("create-variations",), "fields": option_fields}))

class ProductAdmin(DisplayableAdmin):

    list_display = ("admin_thumb", "title", "status", "available", "admin_link")
    list_display_links = ("admin_thumb", "title")
    list_editable = ("status", "available")
    list_filter = ("status", "available", "categories")
    filter_horizontal = ("categories",)
    search_fields = ("title", "content", "categories__title", "variations__sku")
    inlines = (ProductImageAdmin, ProductVariationAdmin)
    form = ProductAdminForm
    fieldsets = product_fieldsets

    def save_model(self, request, obj, form, change):
        """
        Store the product object for creating variations in save_formset.
        """
        super(ProductAdmin, self).save_model(request, obj, form, change)
        self._product = obj

    def save_formset(self, request, form, formset, change):
        """
        Create variations for selected options if they don't exist, manage the 
        default empty variation creating it if no variations exist or removing 
        it if multiple variations exist, and copy the pricing and image fields
        from the default variation to the product.
        """
        super(ProductAdmin, self).save_formset(request, form, formset, change)
        if isinstance(formset, ProductVariationAdminFormset):
            options = dict([(f, request.POST.getlist(f)) for f in option_fields 
                if request.POST.getlist(f)])
            self._product.variations.create_from_options(options)
            self._product.variations.manage_empty()
            self._product.copy_default_variation()


class ProductOptionAdmin(admin.ModelAdmin):
    ordering = ("type", "name")
    list_display = ("type", "name")
    list_display_links = ("type",)
    list_editable = ("name",)
    list_filter = ("type",)
    search_fields = ("type", "name")
    radio_fields = {"type": admin.HORIZONTAL}


class OrderItemInline(admin.TabularInline):
    verbose_name_plural = _("Items")
    model = OrderItem
    extra = 0
    formfield_overrides = {MoneyField: {"widget": MoneyWidget}}


class OrderAdmin(admin.ModelAdmin):
    ordering = ("status", "-id")
    list_display = ("id", "billing_name", "total", "time", "status")
    list_editable = ("status",)
    list_filter = ("status", "time")
    list_display_links = ("id", "billing_name",)
    search_fields = ["id", "status"] + billing_fields + shipping_fields
    date_hierarchy = "time"
    radio_fields = {"status": admin.HORIZONTAL}
    inlines = (OrderItemInline,)
    formfield_overrides = {MoneyField: {"widget": MoneyWidget}}
    fieldsets = (
        (_("Billing details"), {"fields": (tuple(billing_fields),)}),
        (_("Shipping details"), {"fields": (tuple(shipping_fields),)}),
        (None, {"fields": ("additional_instructions", ("shipping_total", 
            "shipping_type"), ("discount_total", "discount_code"), 
            "item_total",("total", "status"))}),
    )


class SaleAdmin(admin.ModelAdmin):
    list_display = ("title", "active", "discount_deduct", "discount_percent", 
        "discount_exact", "valid_from", "valid_to")
    list_editable = ("active", "discount_deduct", "discount_percent", 
        "discount_exact", "valid_from", "valid_to")
    filter_horizontal = ("categories", "products")
    formfield_overrides = {MoneyField: {"widget": MoneyWidget}}
    form = DiscountAdminForm
    fieldsets = (
        (None, {"fields": ("title", "active")}),
        (_("Apply to product and/or products in categories"), 
            {"fields": ("products", "categories")}),
        (_("Reduce unit price by"), 
            {"fields": (("discount_deduct", "discount_percent", 
            "discount_exact"),)}),
        (_("Sale period"), {"fields": (("valid_from", "valid_to"),)}),
    )


class DiscountCodeAdmin(admin.ModelAdmin):
    list_display = ("title", "active", "code", "discount_deduct", 
        "discount_percent", "min_purchase", "free_shipping", "valid_from", 
        "valid_to")
    list_editable = ("active", "code", "discount_deduct", "discount_percent", 
        "min_purchase", "free_shipping", "valid_from", "valid_to")
    filter_horizontal = ("categories", "products")
    formfield_overrides = {MoneyField: {"widget": MoneyWidget}}
    form = DiscountAdminForm
    fieldsets = (
        (None, {"fields": ("title", "active", "code")}),
        (_("Apply to product and/or products in categories"), 
            {"fields": ("products", "categories")}),
        (_("Reduce unit price by"), 
            {"fields": (("discount_deduct", "discount_percent"),)}),
        (None, {"fields": (("min_purchase", "free_shipping"),)}),
        (_("Valid for"), {"fields": (("valid_from", "valid_to"),)}),
    )


admin.site.register(Category, CategoryAdmin)
admin.site.register(Product, ProductAdmin)
admin.site.register(ProductOption, ProductOptionAdmin)
admin.site.register(Order, OrderAdmin)
admin.site.register(Sale, SaleAdmin)
admin.site.register(DiscountCode, DiscountCodeAdmin)




import locale
import hmac
from datetime import datetime, timedelta
try:
    from hashlib import sha512 as digest
except ImportError:
    from md5 import new as digest

from django.core.exceptions import ImproperlyConfigured
from django.core.mail import EmailMultiAlternatives
from django.template import loader, Context
from django.utils.translation import ugettext as _

from mezzanine.conf import settings


def make_choices(choices):
    """
    Zips a list with itself for field choices.
    """
    return zip(choices, choices)


def set_shipping(request, shipping_type, shipping_total):
    """
    Stores the shipping type and total in the session.
    """
    if request.session.get("free_shipping", False):
        shipping_type = _("Free shipping")
        shipping_total = 0
    request.session["shipping_type"] = shipping_type
    request.session["shipping_total"] = shipping_total


def set_cookie(response, name, value, secure=False):
    """
    Sets a cookie that expires in a year.
    """
    expires = datetime.strftime(datetime.utcnow() + 
        timedelta(seconds=365*24*60*60), "%a, %d-%b-%Y %H:%M:%S GMT")
    response.set_cookie(name, value, expires=expires, secure=secure)


def sign(value):
    """
    Returns the hash of the given value, used for signing order key stored in 
    cookie for remembering address fields.
    """
    return hmac.new(settings.SECRET_KEY, value, digest).hexdigest()


def send_mail_template(subject, template, addr_from, addr_to, context=None,
    attachments=None, fail_silently=False):
    """
    Send email rendering text and html versions for the specified template name
    using the context dictionary passed in.
    """
    if context is None:
        context = {}
    if attachments is None:
        attachments = []
    # Allow for a single address to be passed in.
    if not hasattr(addr_to, "__iter__"):
        addr_to = [addr_to]
    # Loads a template passing in vars as context.
    render = lambda type: loader.get_template("%s.%s" % 
        (template, type)).render(Context(context))
    # Create and send email.
    msg = EmailMultiAlternatives(subject, render("txt"), addr_from, addr_to)
    msg.attach_alternative(render("html"), "text/html")
    for attachment in attachments:
        msg.attach(attachment)
    msg.send(fail_silently=fail_silently)


def set_locale():
    """
    Sets the locale for currency formatting.
    """
    try:
        locale.setlocale(locale.LC_MONETARY, settings.SHOP_CURRENCY_LOCALE)
    except:
        raise ImproperlyConfigured(_("Invalid currency locale specified: %s") % 
            settings.SHOP_CURRENCY_LOCALE)




from decimal import Decimal
from operator import mul

from django.core.urlresolvers import reverse
from django.test import TestCase

from mezzanine.conf import settings
from mezzanine.core.models import CONTENT_STATUS_PUBLISHED
from mezzanine.utils.tests import run_pyflakes_for_package

from cartridge.shop.models import Product, ProductOption, ProductVariation
from cartridge.shop.models import Category, Cart


class ShopTests(TestCase):

    def setUp(self):
        published = {"status": CONTENT_STATUS_PUBLISHED}
        self._category = Category.objects.create(**published)
        self._product = Product.objects.create(**published)
        for option_type in settings.SHOP_OPTION_TYPE_CHOICES:
            for i in range(10):
                name = "test%s" % i
                ProductOption.objects.create(type=option_type[0], name=name)

    def test_views(self):
        """
        Test the main shop views for errors.
        """
        # Category.
        response = self.client.get(self._category.get_absolute_url())
        self.assertEqual(response.status_code, 200)
        # Product.
        response = self.client.get(self._product.get_absolute_url())
        self.assertEqual(response.status_code, 200)
        # Cart.
        response = self.client.get(reverse("shop_cart"))
        self.assertEqual(response.status_code, 200)
        # Checkout.
        response = self.client.get(reverse("shop_checkout"))
        self.assertEqual(response.status_code, 200 if not 
            settings.SHOP_CHECKOUT_ACCOUNT_REQUIRED else 302)
        # Login.
        if settings.SHOP_CHECKOUT_ACCOUNT_ENABLED:
            response = self.client.get(reverse("shop_account"))
            self.assertEqual(response.status_code, 200)

    def test_variations(self):
        """
        Test creation of variations from options, and management of empty 
        variations.
        """
        product_options = ProductOption.objects.as_fields()
        total_variations = reduce(mul, [len(values) 
            for values in product_options.values()])
        # Clear variations.
        self._product.variations.all().delete()
        self.assertEqual(self._product.variations.count(), 0)
        # Create single empty variation.
        self._product.variations.manage_empty()
        self.assertEqual(self._product.variations.count(), 1)
        # Create variations from all options.
        self._product.variations.create_from_options(product_options)
        # Should do nothing.
        self._product.variations.create_from_options(product_options)
        # All options plus empty.
        self.assertEqual(self._product.variations.count(), total_variations + 1)
        # Remove empty.
        self._product.variations.manage_empty()
        self.assertEqual(self._product.variations.count(), total_variations)

    def test_stock(self):
        """
        Test stock checking on product variations.
        """
        self._product.variations.all().delete()
        self._product.variations.manage_empty()
        num_in_stock = 10
        variation = self._product.variations.all()[0]
        variation.num_in_stock = num_in_stock
        # Check stock field not in use.
        self.assertTrue(variation.has_stock())
        # Check available and unavailable quantities.
        self.assertTrue(variation.has_stock(num_in_stock))
        self.assertFalse(variation.has_stock(num_in_stock + 1))
        # Check sold out.
        variation = self._product.variations.all()[0]
        variation.num_in_stock = 0
        self.assertFalse(variation.has_stock())

    def test_cart(self):
        """
        Test the cart object and cart add/remove forms.
        """
        product_options = ProductOption.objects.as_fields()
        self._product.variations.all().delete()
        self._product.variations.create_from_options(product_options)
        price = Decimal("20")
        num_in_stock = 5
        variation = self._product.variations.all()[0]
        variation.unit_price = price
        variation.num_in_stock = num_in_stock * 2
        variation.save()

        # Test initial cart.
        cart = Cart.objects.from_request(self.client)
        self.assertFalse(cart.has_items())
        self.assertEqual(cart.total_quantity(), 0)
        self.assertEqual(cart.total_price(), Decimal("0"))

        # Add quantity and check stock levels / cart totals.
        data = dict(zip([field.name for field in 
            ProductVariation.option_fields()], variation.options()))
        data["quantity"] = num_in_stock
        self.client.post(self._product.get_absolute_url(), data)
        cart = Cart.objects.from_request(self.client)
        variation = self._product.variations.all()[0]
        self.assertTrue(variation.has_stock(num_in_stock))
        self.assertFalse(variation.has_stock(num_in_stock * 2))
        self.assertTrue(cart.has_items())
        self.assertEqual(cart.total_quantity(), num_in_stock)
        self.assertEqual(cart.total_price(), price * num_in_stock)

        # Add remaining quantity and check again.
        self.client.post(self._product.get_absolute_url(), data)
        cart = Cart.objects.from_request(self.client)
        variation = self._product.variations.all()[0]
        self.assertFalse(variation.has_stock())
        self.assertTrue(cart.has_items())
        self.assertEqual(cart.total_quantity(), num_in_stock * 2)
        self.assertEqual(cart.total_price(), price * num_in_stock * 2)

        # Remove from cart.
        for item in cart:
            self.client.post(reverse("shop_cart"), {"item_id": item.id})
        cart = Cart.objects.from_request(self.client)
        variation = self._product.variations.all()[0]
        self.assertTrue(variation.has_stock(num_in_stock * 2))
        self.assertFalse(cart.has_items())
        self.assertEqual(cart.total_quantity(), 0)
        self.assertEqual(cart.total_price(), Decimal("0"))

    def test_with_pyflakes(self):
        """
        Run pyflakes across the code base to check for potential errors.
        """
        ignore = ("redefinition of unused 'digest'", 
                  "redefinition of unused 'info'")
        warnings = run_pyflakes_for_package("cartridge", extra_ignore=ignore)
        if warnings:
            warnings.insert(0, "pyflakes warnings:")
            self.fail("\n".join(warnings))
